@@ -1,10 +1,10 @@
 ;;; dired-rsync.el --- Allow rsync from dired buffers -*- lexical-binding: t -*-
 ;;
-;; Copyright (C) 2018 Alex Bennée
+;; Copyright (C) 2018, 2019, 2020 Alex Bennée
 ;;
 ;; Author: Alex Bennée <alex@bennee.com>
 ;; Maintainer: Alex Bennée <alex@bennee.com>
-;; Version: 0.5
+;; Version: 0.6
 ;; Package-Requires: ((s "1.12.0") (dash "2.0.0") (emacs "24"))
 ;; Homepage: https://github.com/stsquad/dired-rsync
 ;;
@@ -88,8 +88,8 @@ It is run in the context of the failed process buffer."
   :group 'dired-rsync)
 
 ;; Internal variables
-(defvar dired-rsync-job-count 0
-  "Count of running rsync jobs.")
+(defvar dired-rsync-proc-buffer-prefix "*rsync"
+  "Prefix for process buffers.")
 
 (defvar dired-rsync-modeline-status
   ""
@@ -103,14 +103,26 @@ It is run in the context of the failed process buffer."
   (rx (** 1 3 digit) "%")
   "A regex to extract the % complete from a file.")
 
+(defvar dired-remote-portfwd
+  "ssh -p %d -o StrictHostKeyChecking=no -o UserKnownHostsFile=/dev/null"
+  "An explicit ssh command for rsync to use port forwarded proxy.
+The string is treated as a format string where %d is replaced with the
+results of `dired-rsync--get-remote-port'.")
+
 ;; Helpers
+(defun dired-rsync--get-remote-port ()
+  "Return the remote port we shall use for the reverse port-forward."
+  (+ 50000 (length (dired-rsync--get-active-buffers))))
+
+(defun dired-rsync--get-remote-portfwd ()
+  (format dired-remote-portfwd (dired-rsync--get-remote-port)))
 
 (defun dired-rsync--quote-and-maybe-convert-from-tramp (file-or-path)
   "Reformat a tramp FILE-OR-PATH to one usable for rsync."
   (if (tramp-tramp-file-p file-or-path)
-      ;; tramp format is /method:remote:path
-      (let ((parts (s-split ":" file-or-path)))
-        (format "%s:\"%s\"" (nth 1 parts) (shell-quote-argument (nth 2 parts))))
+      (with-parsed-tramp-file-name file-or-path tfop
+        (format "%s%s:\"%s\"" (if tfop-user (format "%s@" tfop-user) "") tfop-host
+                (shell-quote-argument tfop-localname)))
     (shell-quote-argument file-or-path)))
 
 (defun dired-rsync--extract-host-from-tramp (file-or-path &optional split-user)
@@ -118,27 +130,45 @@ It is run in the context of the failed process buffer."
 
 It SPLIT-USER is set we remove the user@ part as well.  We assume
 hosts don't need quoting."
-  (let ((parts (s-split ":" file-or-path)))
-    (let ((host (nth 1 parts)))
-      (if (and split-user (s-contains? "@" host))
-          (nth 1 (s-split "@" host))
-        host))))
-
+  (with-parsed-tramp-file-name file-or-path tfop
+    (if (or split-user (not tfop-user))
+        tfop-host
+      (format "%s@%s" tfop-user tfop-host))))
 
+;; This is tricky for remote-to-remote because we may have an implied
+;; user from the local config which isn't available on the remote
+;; .ssh/config
 (defun dired-rsync--extract-user-from-tramp (file-or-path)
   "Extract the username part of a tramp FILE-OR-PATH."
-  (when (s-contains? "@" file-or-path)
-    (nth 1 (s-split ":" (nth 0 (s-split "@" file-or-path))))))
+  (with-parsed-tramp-file-name file-or-path tfop
+    (or tfop-user
+        ; somehow extract .ssh/config user for tfop-host
+        (getenv "USER"))))
 
 
 (defun dired-rsync--extract-paths-from-tramp (files)
   "Extract the path part of a tramp FILES and quote it."
   (--map
-   (let ((parts (s-split ":" it)))
-     (shell-quote-argument (nth 2 parts)))
+   (with-parsed-tramp-file-name it tfop
+     (shell-quote-argument tfop-localname))
    files))
 
 
+;; Count active buffers
+(defun dired-rsync--get-proc-buffers ()
+  "Return all dired-rsync process buffers."
+  (--filter
+   (and (s-starts-with? dired-rsync-proc-buffer-prefix (buffer-name it))
+        (get-buffer-process it))
+   (buffer-list)))
+
+(defun dired-rsync--get-active-buffers ()
+  "Return all currently active process buffers"
+  (--filter
+   (process-live-p (get-buffer-process it))
+   (dired-rsync--get-proc-buffers)))
+
+
 ;; Update status with count/speed
 (defun dired-rsync--update-modeline (&optional err ind)
   "Update the modeline, optionally with `ERR' or `IND'.
@@ -147,20 +177,25 @@ hosts don't need quoting."
 alternative indication (such as a percentage completion).  If
 neither is set we simply display the current number of jobs."
   (force-mode-line-update)
-  (setq mode-line-process
-        (setq dired-rsync-modeline-status
-              (cond
-               ;; error has occurred
-               (err (propertize
-                     (format " R:%d %s!!" dired-rsync-job-count err)
-                     'font-lock-face '(:foreground "red")))
-               ;; we still have jobs but no error
-               ((> dired-rsync-job-count 1)
-                (format " R:%d" dired-rsync-job-count))
-               ((> dired-rsync-job-count 0)
-                (format " R:%s" (or ind dired-rsync-job-count)))
-               ;; nothing going on
-               (t nil)))))
+  (let ((job-count (length (dired-rsync--get-active-buffers))))
+    (setq dired-rsync-modeline-status
+          (cond
+           ;; error has occurred
+           (err (propertize
+                 (format " R:%d %s!!" job-count err)
+                 'font-lock-face '(:foreground "red")))
+           ;; we still have jobs but no error
+           ((> job-count 1)
+            (format " R:%d" job-count))
+           ((> job-count 0)
+            (format " R:%s" (or ind job-count)))
+           ;; Any stale?
+           ((dired-rsync--get-proc-buffers)
+            (propertize
+                 " R:hung :-("
+                 'font-lock-face '(:foreground "red")))
+           ;; nothing going on
+           (t nil)))))
 
 ;;
 ;; Running rsync: We need to take care of a couple of things here. We
@@ -179,13 +214,11 @@ This gets called whenever the inferior `PROC' changes state as
     (when (s-starts-with-p "finished" desc)
       ;; clean-up finished tasks
       (let ((dired-buf (plist-get details ':dired-buffer)))
-        (when dired-rsync-unmark-on-completion
+        (when (and dired-rsync-unmark-on-completion
+                   (buffer-live-p dired-buf))
           (with-current-buffer dired-buf
             (dired-unmark-all-marks)))
         (kill-buffer proc-buf)))
-    ;; clean-up data left from dead/finished processes
-    (when (not (process-live-p proc))
-      (setq dired-rsync-job-count (1- dired-rsync-job-count)))
     (dired-rsync--update-modeline)
     ;; If we still have a process buffer things didn't end well
     (when (and (not (process-live-p proc))
@@ -212,8 +245,7 @@ dired-buffer modeline."
       (process-send-string proc (concat (read-passwd string) "\n")))
     ;; update if anything to report
     (when (or err indicator)
-      (with-current-buffer (plist-get details ':dired-buffer)
-        (dired-rsync--update-modeline err indicator))))
+      (dired-rsync--update-modeline err indicator)))
 
   ;; update the process buffer (we could just drop?)
   (let ((old-process-mark (process-mark proc)))
@@ -231,7 +263,7 @@ dired-buffer modeline."
 
 (defun dired-rsync--do-run (command details)
   "Run rsync COMMAND in a unique buffer, passing DETAILS to sentinel."
-  (let* ((buf (format "*rsync @ %s" (current-time-string)))
+  (let* ((buf (format "%s @ %s" dired-rsync-proc-buffer-prefix (current-time-string)))
          (proc (start-process-shell-command "*rsync*" buf command)))
     (set-process-sentinel
      proc
@@ -241,7 +273,6 @@ dired-buffer modeline."
      proc
      #'(lambda (proc string)
          (dired-rsync--filter proc string details)))
-    (setq dired-rsync-job-count (1+ dired-rsync-job-count))
     (dired-rsync--update-modeline)))
 
 (defun dired-rsync--remote-to-from-local-cmd (sfiles dest)
@@ -270,13 +301,14 @@ destination.  This requires ssh'ing to the source and running the rsync
 there."
   (s-join " " (-flatten
                (list "ssh" "-A"
-                     "-R" (format "localhost:50000:%s:22" dhost)
+                     "-R" (format "localhost:%d:%s:22"
+                                  (dired-rsync--get-remote-port) dhost)
                      shost
                      (format
-                      "'%s %s -e \"%s\" %s %s@localhost:%s'"
+                      "\"%s %s -e \\\"%s\\\" %s %s@localhost:%s\""
                       dired-rsync-command
                       dired-rsync-options
-                      "ssh -p 50000 -o StrictHostKeyChecking=no -o UserKnownHostsFile=/dev/null"
+                      (dired-rsync--get-remote-portfwd)
                       (s-join " " sfiles)
                       duser
                       dpath)))))
@@ -294,7 +326,7 @@ the copy is running.  It also handles both source and destinations on
 ssh/scp tramp connections."
   ;; Interactively grab dest if not called with
   (interactive
-   (list (read-file-name "rsync to:" (dired-dwim-target-directory)
+   (list (read-file-name "rsync to: " (dired-dwim-target-directory)
                          nil nil nil 'file-directory-p)))
 
   (setq dest (expand-file-name dest))
